;----------------------------------------------------------------
; CONSTANTS:
;----------------------------------------------------------------

PRGCOUNT = 2
; 16kb = 1
; 32kb = 2

CHRCOUNT = 1

MIRRORING = %0000
; Horizontal =  %0000
; Vertical =    %0001
; Four Screen = %1000

VAR.GameState.Initialize EQU #$00
VAR.GameState.Construct EQU #$01
VAR.GameState.Update EQU #$02
VAR.GameState.Deconstruct EQU #$03

VAR.NMIState.DisableRendering EQU #$00
VAR.NMIState.LoadPalettes EQU #$01
VAR.NMIState.EnableRendering EQU #$02
VAR.NMIState.Wait EQU #$03
VAR.NMIState.Normal EQU #$04


;----------------------------------------------------------------
; VARIABLES:
;----------------------------------------------------------------

.enum $0000

VAR.Scene .dsb 1
VAR.Scene.Next .dsb 1
VAR.Frame .dsb 1
VAR.GameState .dsb 1
VAR.NMIState .dsb 1

PTR.Background.Nametable .dsw 1
PTR.Background.Attributes .dsw 1
PTR.Background.Palette .dsw 1

vAR.Sprites.Size .dsb 1

PTR.Sprites .dsw 1
PTR.Sprites.Palette .dsw 1

VAR.PPUStack.High .dsb 1
VAR.PPUStack.Low .dsb 1
VAR.PPUStack.Value .dsb 1

; bit:       7     6     5     4     3     2     1     0
; button:    A     B   select start  up   down  left right

VAR.Controller1  .dsb 1
VAR.Controller1.Old .dsb 1
VAR.Controller1.Pressed .dsb 1
VAR.Controller1.Released .dsb 1

VAR.Controller2 .dsb 1
VAR.Controller2.Old .dsb 1
VAR.Controller2.Pressed .dsb 1
VAR.Controller2.Released .dsb 1

VAR.Scene0.Menu.SpriteIndex .dsb 1

.ende

.enum $0200
VAR.PPUStack.Size .dsb 1
BUF.PPUStack.High .dsb 80
BUF.PPUStack.Low .dsb 80
BUF.PPUStack.Value .dsb 80
.ende

.enum $0300
BUF.Sprites .dsb 256
.ende


;----------------------------------------------------------------
; iNES header
;----------------------------------------------------------------

   .db "NES", $1a 			;identification of the iNES header
   .db PRGCOUNT 			;number of 16KB PRG-ROM pages
   .db CHRCOUNT				;number of 8KB CHR-ROM pages
   .db $00|MIRRORING 			;mapper 0 and mirroring
   .dsb 9, $00 				;clear the remaining bytes

;----------------------------------------------------------------
; CODE: Main Loop
;----------------------------------------------------------------

.base $10000-(PRGCOUNT*$4000)

LBL.RESET:
  SEI          ; disable IRQs
  CLD          ; disable decimal mode

  LDX #$40
  STX $4017    ; disable APU frame IRQ
  
  LDX #$00
  STX $2000    ; disable NMI
  STX $2001    ; disable rendering
  STX $4010    ; disable DMC IRQs

  STA VAR.Controller1.Old
  STA VAR.Controller2.Old

  LDA VAR.NMIState.Wait
  STA VAR.NMIState

  BIT $2002

  JSR LBL.VBlankWait

  JSR LBL.ClearRAM  
  JSR LBL.ClearSpriteBuffer

  JSR LBL.VBlankWait

  LDA #$00		; Start Game at Scene 0
  STA VAR.Scene

  LDA VAR.GameState.Initialize
  STA VAR.GameState

  LDA #%10010000   ; enable NMI, sprites from Pattern Table 0, background from Pattern Table 1
  STA $2000

LBL.Main 

  JSR LBL.GameState.Launcher
  JMP LBL.Main

;--------------------------------------------------------------------
;CODE:  Main Subroutines
;--------------------------------------------------------------------

LBL.ClearRAM:
  LDX #$00
  LDA #$00

LBL.ClearRAM.Loop:
  STA $0000, x
  STA $0200, x
  STA $0400, x
  STA $0500, x
  STA $0600, x
  STA $0700, x
  INX
  BNE LBL.ClearRAM.Loop
  RTS

;--------------------------------------------------------------------

LBL.ClearSpriteBuffer:
  LDX #$00
  LDA #$FF

LBL.ClearSpriteBuffer.Loop:
  STA $0300, x
  INX
  BNE LBL.ClearSpriteBuffer.Loop
  RTS

;--------------------------------------------------------------------

LBL.VBlankWait:
  BIT $2002
  BPL LBL.VBlankWait
  RTS
  
;--------------------------------------------------------------------

LBL.GameState.Initialize:
  JSR LBL.Scene.Initialize.Launcher
  LDA VAR.GameState.Construct
  STA VAR.GameState
  RTS

;--------------------------------------------------------------------

LBL.GameState.Launcher:
  LDA VAR.GameState
  ASL
  TAX
  LDA TBL.GameState + 1, x
  PHA
  LDA TBL.GameState, x
  PHA
  RTS

;--------------------------------------------------------------------

LBL.Scene.Initialize.Launcher:
  LDA VAR.Scene
  ASL
  TAX
  LDA TBL.Scene.Initialize + 1, x
  PHA
  LDA TBL.Scene.Initialize, x
  PHA
  RTS

;--------------------------------------------------------------------

LBL.GameState.Construct:
  LDA VAR.NMIState.DisableRendering
  STA VAR.NMIState
  JSR LBL.WaitForNMIStateChange

  JSR LBL.LoadBackground

  JSR LBL.ClearSpriteBuffer
  JSR LBL.LoadSprites

  LDA VAR.NMIState.LoadPalettes
  STA VAR.NMIState
  JSR LBL.WaitForNMIStateChange

  LDA VAR.NMIState.EnableRendering
  STA VAR.NMIState
  JSR LBL.WaitForNMIStateChange

  LDA #$00
  STA VAR.Frame

  LDA VAR.NMIState.Normal
  STA VAR.NMIState
  
  LDA VAR.GameState.Update
  STA VAR.GameState

  LDA #$00
  STA VAR.Controller1.Old

  LDA #$00
  STA VAR.Controller2.Old

  RTS

;--------------------------------------------------------------------

LBL.WaitForNMIStateChange:
  LDA VAR.NMIState
LBL.WaitForNMIStateChange.Loop:
  CMP VAR.NMIState
  BEQ LBL.WaitForNMIStateChange.Loop
  RTS

;--------------------------------------------------------------------

LBL.WaitForFrameChange:
  LDA VAR.Frame
LBL.WaitForFrameChange.Loop:
  CMP VAR.Frame
  BEQ LBL.WaitForFrameChange.Loop
  RTS

;--------------------------------------------------------------------

LBL.GameState.Update:
  JSR LBL.WaitForFrameChange
  JSR LBL.ReadControllers
  JSR LBL.Scene.Update.Launcher
  RTS

;--------------------------------------------------------------------

LBL.GameState.Deconstruct:
  JSR LBL.Scene.Deconstruct.Launcher
  LDA VAR.Scene.Next
  STA VAR.Scene
  LDA VAR.GameState.Initialize
  STA VAR.GameState
  RTS

;--------------------------------------------------------------------

LBL.Scene.Update.Launcher:
  LDA VAR.Scene
  ASL
  TAX
  LDA TBL.Scene.Update + 1, x
  PHA
  LDA TBL.Scene.Update, x
  PHA
  RTS

;--------------------------------------------------------------------

LBL.Scene.Deconstruct.Launcher:
  LDA VAR.Scene
  ASL
  TAX
  LDA TBL.Scene.Deconstruct + 1, x
  PHA
  LDA TBL.Scene.Deconstruct, x
  PHA
  RTS
  
;--------------------------------------------------------------------



LBL.ReadControllers:
  LDA #$01
  STA $4016
  LDA #$00
  STA $4016
  LDX #$08

  LDA VAR.Controller1;
  STA VAR.Controller1.Old;
  
  LDA VAR.Controller2;
  STA VAR.Controller2.Old;

LBL.ReadController1.Loop:
  LDA $4016
  LSR A
  ROL VAR.Controller1
  DEX
  BNE LBL.ReadController1.Loop


  LDX #$08

LBL.ReadController2.Loop:
  LDA $4017
  LSR A
  ROL VAR.Controller2
  DEX
  BNE LBL.ReadController2.Loop

  LDA VAR.Controller1
  EOR #$FF
  AND VAR.Controller1.Old
  STA VAR.Controller1.Released

  LDA VAR.Controller1.Old
  EOR #$FF
  AND VAR.Controller1
  STA VAR.Controller1.Pressed

  LDA VAR.Controller2
  EOR #$FF
  AND VAR.Controller2.Old
  STA VAR.Controller2.Released

  LDA VAR.Controller2.Old
  EOR #$FF
  AND VAR.Controller2
  STA VAR.Controller2.Pressed

  RTS

;--------------------------------------------------------------------

LBL.LoadBackground:
  LDA #$20
  STA $2006
  LDY #$00
  STY $2006

  LDX #$04
  LDY #$00

LBL.LoadBackground.Nametable.Loop:
  LDA (PTR.Background.Nametable), y
  STA $2007
  INY
  BNE LBL.LoadBackground.Nametable.Loop

  INC PTR.Background.Nametable + 1;
  DEX
  BNE LBL.LoadBackground.Nametable.Loop
  RTS


;--------------------------------------------------------------------


LBL.LoadSprites:
  LDY #$00
  CPY vAR.Sprites.Size
  BEQ LBL.LoadSprites.Exit

LBL.LoadSprites.Loop:
  LDA (PTR.Sprites), Y
  STA BUF.Sprites, Y
  INY
  CPY vAR.Sprites.Size
  BNE LBL.LoadSprites.Loop

LBL.LoadSprites.Exit:  
  RTS

;--------------------------------------------------------------------

LBL.PushPPUStack:

  INC VAR.PPUStack.Size
  LDX VAR.PPUStack.Size

  LDA VAR.PPUStack.High;
  STA BUF.PPUStack.High, x;

  LDA VAR.PPUStack.Low;
  STA BUF.PPUStack.Low, x;

  LDA VAR.PPUStack.Value;
  STA BUF.PPUStack.Value, x;

  RTS

;--------------------------------------------------------------------
;CODE:  NMI
;--------------------------------------------------------------------

LBL.NMI:
  PHA
  TXA
  PHA
  TYA
  PHA


  JSR LBL.NMIState.Launcher

  PLA
  TAY
  PLA
  TAX
  PLA

  RTI

;--------------------------------------------------------------------
;CODE:  NMI Subroutines
;--------------------------------------------------------------------

LBL.NMIState.Launcher:
  LDA VAR.NMIState
  ASL
  TAX
  LDA TBL.NMIState + 1, x
  PHA
  LDA TBL.NMIState, x
  PHA
  RTS

;--------------------------------------------------------------------

LBL.NMI.DisableRendering:
  LDA #$00	;Disable Rendering
  STA $2001	;Disable Rendering
  LDA VAR.NMIState.Wait
  STA VAR.NMIState
  RTS

;--------------------------------------------------------------------

LBL.NMI.LoadPalettes:

  LDA $2002 ; reset High/Low Bit in PPU
  LDA #$3F
  STA $2006
  LDA #$00
  STA $2006
  LDY #$00

LBL.NMI.LoadPalettes.BackgroundLoop:
  LDA (PTR.Background.Palette), Y
  STA $2007
  INY
  CPY #$10
  BNE LBL.NMI.LoadPalettes.BackgroundLoop

  LDY #$00

LBL.NMI.LoadPalettes.SpriteLoop:
  LDA (PTR.Sprites.Palette), Y
  STA $2007
  INY
  CPY #$10
  BNE LBL.NMI.LoadPalettes.SpriteLoop

  LDA VAR.NMIState.Wait
  STA VAR.NMIState

  RTS

;--------------------------------------------------------------------

LBL.NMI.EnableRendering:
  LDA #%00011110
  STA $2001
  LDA VAR.NMIState.Wait
  STA VAR.NMIState
  RTS

;--------------------------------------------------------------------

LBL.NMI.Wait:
  RTS

;--------------------------------------------------------------------

LBL.NMI.Normal:
  JSR LBL.CopySprites
  JSR LBL.BurnStack
  INC VAR.Frame
  RTS

;--------------------------------------------------------------------

LBL.CopySprites:
  LDA #$00
  STA $2003
  LDA #>BUF.Sprites
  STA $4014
  RTS

;--------------------------------------------------------------------

LBL.BurnStack:
  LDA $2002
  LDX vAR.Sprites.Size
  BEQ LBL.Burnstack.Exit

LBL.BurnStack.Loop:
  LDA VAR.PPUStack.High,x
  STA $2006
  LDA VAR.PPUStack.Low,x
  STA $2006
  LDA VAR.PPUStack.Value,x
  STA $2007
  DEX
  BNE LBL.BurnStack.Loop

LBL.Burnstack.Exit:
  RTS

;----------------------------------------------------------------
; Scene0
;----------------------------------------------------------------

LBL.Scene0.Initialize
  
  LDA #<DAT.Scene0.Background.Nametable
  STA PTR.Background.Nametable
  LDA #>DAT.Scene0.Background.Nametable
  STA PTR.Background.Nametable + 1

  LDA #<DAT.Scene0.Background.Palette
  STA PTR.Background.Palette
  LDA #>DAT.Scene0.Background.Palette
  STA PTR.Background.Palette + 1

  LDA #<DAT.Scene0.Sprites
  STA PTR.Sprites
  LDA #>DAT.Scene0.Sprites
  STA PTR.Sprites + 1

  LDA #<DAT.Scene0.Sprites.Palette
  STA PTR.Sprites.Palette
  LDA #>DAT.Scene0.Sprites.Palette
  STA PTR.Sprites.Palette + 1

  LDA #$00 + VAR.Scene0.Sprites.Count*4
  STA vAR.Sprites.Size

  LDA #$00
  STA VAR.Scene0.Menu.SpriteIndex

  RTS

LBL.Scene0.Update:

  LDA VAR.Controller1.Pressed

LBL.Scene0.Pressed.Up:
  LDA VAR.Controller1.Pressed
  CMP #%00001000
  BNE LBL.Scene0.Pressed.Down
    DEC VAR.Scene0.Menu.SpriteIndex
    LDA VAR.Scene0.Menu.SpriteIndex
    CMP #$FF
    BNE LBL.Scene0.Update.Exit
      LDA #$00 + VAR.Scene0.Menu.Count - 1
      STA VAR.Scene0.Menu.SpriteIndex
      JMP LBL.Scene0.Update.Exit

LBL.Scene0.Pressed.Down:
  LDA VAR.Controller1.Pressed
  CMP #%00000100
  BNE LBL.Scene0.Pressed.Start
    INC VAR.Scene0.Menu.SpriteIndex
    LDA VAR.Scene0.Menu.SpriteIndex
    CMP #$00 + VAR.Scene0.Menu.Count
    BNE LBL.Scene0.Update.Exit
      LDA #$00
      STA VAR.Scene0.Menu.SpriteIndex
      JMP LBL.Scene0.Update.Exit


LBL.Scene0.Pressed.Start:
  LDA VAR.Controller1.Pressed
  CMP  #%00100000
  BNE  LBL.Scene0.Update.Exit
     LDX VAR.Scene0.Menu.SpriteIndex
     LDA DAT.Scene0.NextSceneSelected, X
     STA VAR.Scene.Next
     LDA VAR.GameState.Deconstruct
     STA VAR.GameState
     RTS

LBL.Scene0.Update.Exit:
  LDX VAR.Scene0.Menu.SpriteIndex
  LDA DAT.Scene0.Menu.SpritePositionY, X
  STA BUF.Sprites
  RTS


LBL.Scene0.Deconstruct:
  RTS

;----------------------------------------------------------------
; Scene1
;----------------------------------------------------------------

LBL.Scene1.Initialize
  
  LDA #<DAT.Scene1.Background.Nametable
  STA PTR.Background.Nametable
  LDA #>DAT.Scene1.Background.Nametable
  STA PTR.Background.Nametable + 1

  LDA #<DAT.Scene1.Background.Palette
  STA PTR.Background.Palette
  LDA #>DAT.Scene1.Background.Palette
  STA PTR.Background.Palette + 1

  LDA #<DAT.Scene1.Sprites
  STA PTR.Sprites
  LDA #>DAT.Scene1.Sprites
  STA PTR.Sprites + 1

  LDA #<DAT.Scene1.Sprites.Palette
  STA PTR.Sprites.Palette
  LDA #>DAT.Scene1.Sprites.Palette
  STA PTR.Sprites.Palette + 1

  LDA #$00 + VAR.Scene1.Sprites.Count*4
  STA vAR.Sprites.Size

  RTS

LBL.Scene1.Update:
  RTS

LBL.Scene1.Deconstruct:
  RTS


;----------------------------------------------------------------
; JUMP TABLE(S)
;----------------------------------------------------------------

.org $DD00

TBL.GameState:
  .dw LBL.GameState.Initialize - 1
  .dw LBL.GameState.Construct - 1
  .dw LBL.GameState.Update - 1
  .dw LBL.GameState.Deconstruct - 1

TBL.NMIState:
  .dw LBL.NMI.DisableRendering - 1
  .dw LBL.NMI.LoadPalettes - 1
  .dw LBL.NMI.EnableRendering - 1
  .dw LBL.NMI.Wait - 1
  .dw LBL.NMI.Normal - 1

TBL.Scene.Initialize:
  .dw LBL.Scene0.Initialize - 1
  .dw LBL.Scene1.Initialize - 1

TBL.Scene.Update:
  .dw LBL.Scene0.Update - 1
  .dw LBL.Scene1.Update - 1

TBL.Scene.Deconstruct:
  .dw LBL.Scene0.Deconstruct - 1
  .dw LBL.Scene1.Deconstruct - 1

;----------------------------------------------------------------
;SCENE 0 : DATA
;----------------------------------------------------------------

.org $E000    ;;align the background data so the lower address is $00

DAT.Scene0.Background.Nametable:
  .incbin "Nametable_Scene0.nam"

DAT.Scene0.Background.Attributes:
  .incbin "Attributes_Scene0.atr"
 
DAT.Scene0.Background.Palette:
  .db $0f,$30,$10,$00,$0f,$01,$21,$31,$0f,$06,$16,$26,$0f,$09,$19,$29


DAT.Scene0.Sprites.Palette:
  .db $0f,$30,$10,$00,$0f,$01,$21,$31,$0f,$06,$16,$26,$0f,$09,$19,$29


DAT.Scene0.Sprites:
  ;vert tile attr horiz
  .db $0F, $64, $00, $04   ;sprite 0

DAT.Scene0.Menu.SpritePositionY:
  .db $0F, $17

DAT.Scene0.NextSceneSelected:
  .db $01, $01

VAR.Scene0.Sprites.Count = 1
VAR.Scene0.Menu.Count = 2

;----------------------------------------------------------------
;SCENE 1 : DATA
;----------------------------------------------------------------

.org $F000    ;;align the background data so the lower address is $00

DAT.Scene1.Background.Nametable:
  .incbin "Nametable_Scene1.nam"

DAT.Scene1.Background.Attributes:
  .incbin "Attributes_Scene1.atr"
 
DAT.Scene1.Background.Palette:
  .db $0f,$30,$10,$00,$0f,$01,$21,$31,$0f,$06,$16,$26,$0f,$09,$19,$29


DAT.Scene1.Sprites.Palette:
  .db $0f,$30,$10,$00,$0f,$01,$21,$31,$0f,$06,$16,$26,$0f,$09,$19,$29


DAT.Scene1.Sprites:
  ;vert tile attr horiz
  .db $00, $00, $00, $00   ;sprite 0

VAR.Scene1.Sprites.Count = 0

;----------------------------------------------------------------
; interrupt vectors
;----------------------------------------------------------------

   .org $FFFA

   .dw LBL.NMI
   .dw LBL.RESET
   .dw 0

;----------------------------------------------------------------
; CHR-ROM bank
;----------------------------------------------------------------

   .incbin "mario.chr"